/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: CmdDisplay.h,v $
 *	$Author: dalke $	$Locker:  $		$State: Exp $
 *	$Revision: 1.11 $	$Date: 1997/03/24 17:21:32 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *   These are the Commands that control the various aspects
 * of the display, like, the clipping planes, eye separation, etc.
 *
 ***************************************************************************/
#ifndef CMDDISPLAY_H
#define CMDDISPLAY_H

#include "Command.h"
#include "DisplayDevice.h"
#include "utilities.h"

// These are the Command:: enums used by this file:
//  DISP_RESHAPE, DISP_RESETVIEW, DISP_STEREO, DISP_PROJ, DISP_EYESEP, 
//  DISP_FOCALLEN, DISP_LIGHT_ON, DISP_LIGHT_HL, DISP_LIGHT_ROT, 
//  DISP_MATERIALS_CHANGE, DISP_CLIP, DISP_DEPTHCUE, DISP_ANTIALIAS,
//  DISP_DETAIL, DISP_ALTDETAIL, DISP_DEPTHSORT, DISP_UPDATE, CMD_AXES, CMD_STAGE

// text callback routines for 'display', 'light', 'axes, and 'stage'
typedef struct Tcl_Interp;
extern int text_cmd_display(int, char **, CommandQueue *, int, Tcl_Interp *);
extern int text_cmd_light(int, char **, CommandQueue *, int, Tcl_Interp *);
extern int text_cmd_axes(int, char **, CommandQueue *, int, Tcl_Interp *);
extern int text_cmd_stage(int, char **, CommandQueue *, int, Tcl_Interp *);


//////////////////// update the display, or toggle updating on/off
class CmdDisplayUpdate : public Command {
private:
  int newUpdateValue;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayUpdate(char *, int newUIid = (-1));
};


//////////////////// reshape the display
class CmdDisplayReshape : public Command {
protected:
  virtual int do_execute(void);
public:
  CmdDisplayReshape(int newUIid = (-1));
};


///////////////// reset the current view for the current scene
class CmdResetView : public Command {
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdResetView(int newUIid=(-1)) ;
};


//////////////////// set stereo mode of display
class CmdDisplayStereo : public Command {
private:
  int stmode;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayStereo(int newstmode, int newUIid = (-1));
};


//////////////// set the projection to either perspective or orthographic
class CmdDisplayProj : public Command {
public:
  CmdDisplayProj(DisplayDevice::Projection newproj, int newUIid = (-1));
private:
  DisplayDevice::Projection proj;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
};


//////////////////// set eye separation of display
class CmdDisplayEyesep : public Command {
private:
  float sep;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayEyesep(float newsep, int newUIid = (-1));
};


/////////////////// change the rendering detail
class CmdDisplayDetail : public Command {
private:
  int my_level;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayDetail(int new_level, int newUIid = (-1));
};

/////////////////// change the alternate rendering detail
class CmdDisplayAltDetail : public Command {
private:
  int my_level;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayAltDetail(int new_level, int newUIid = (-1));
};

/////////////////// change the depth-sorting parameter
class CmdDisplayDepthSort : public Command {
private:
  int my_sort;
protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayDepthSort(int new_sort, int newUIid = (-1));
};

//////////////////// set focal length of display
class CmdDisplayFocallen : public Command {
private:
  float flen;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayFocallen(float newlen, int newUIid = (-1));
};


//////////////////// set screen height value
class CmdDisplayScreenHeight : public Command {
private:
  float val;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayScreenHeight(float newval, int newUIid = (-1));
};


//////////////////// set distance to screen from origin
class CmdDisplayScreenDistance : public Command {
private:
  float val;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayScreenDistance(float newval, int newUIid = (-1));
};


//////////////////// turn on/off antialiasing
class CmdDisplayAAOn : public Command {
private:
  int onoff;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayAAOn(int turnon, int newUIid = (-1));
};


//////////////////// turn on/off depth cueing
class CmdDisplayDepthcueOn : public Command {
private:
  int onoff;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayDepthcueOn(int turnon, int newUIid = (-1));
};


/////////////////////  clipping plane controls
// This handles the whole range of clipping plane options
//  There are derived classes so you won't have to have the funky flags
// to change {near,fixed} clipping plane {to a fixed,by a relative} amount
// or not
class CmdDisplayClip : public Command {
private:
  int changenear;		// near (T) or far (F) clip plane
  int setval;			// absolute (T) or relative (F) change
  float amount;			// how much to change
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayClip(int ischangenear, int issetval, float newamt, int UIid = -1);
};


//  Change the near clipping plane to a fixed value
// leaving everything else up to the base class :)
class CmdDisplayClipNear : public CmdDisplayClip {
 public:
   CmdDisplayClipNear( float near, int UIid = -1) : 
        CmdDisplayClip(TRUE, TRUE, near, UIid) {
   }
};

//  Change the near clipping plane to by a relative value
class CmdDisplayClipNearRel : public CmdDisplayClip {
 public:
   CmdDisplayClipNearRel( float near, int UIid = -1) : 
         CmdDisplayClip(TRUE, FALSE, near, UIid) {
   }
};

//  Change the far clipping plane to a fixed value
class CmdDisplayClipFar : public CmdDisplayClip {
 public:
   CmdDisplayClipFar( float far, int UIid = -1) : 
          CmdDisplayClip(FALSE, TRUE, far, UIid) {
   }
};

//  Change the far clipping plane to by a relative value
class CmdDisplayClipFarRel : public CmdDisplayClip {
 public:
   CmdDisplayClipFarRel( float far, int UIid = -1) : 
          CmdDisplayClip(FALSE, FALSE, far, UIid) {
   }
};


////////////////////  change the axes location
class CmdDisplayAxes : public Command {
private:
  int newpos;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayAxes(int npos, int newUIid = (-1));
};
    

////////////////////  change the stage location
class CmdDisplayStage : public Command {
public:
  // stage settings this can affect
  enum Settings { LOCATION, PANELS };

private:
  int newpos;
  int settingToChange;

protected:
  virtual int do_execute(void);
  virtual void create_text(void);
public:
  CmdDisplayStage(Settings toset, int npos, int newUIid = (-1));
};


//////////////////// turn on/off the Nth light
class CmdDisplayLightOn : public Command {
private:
  int n, onoff;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayLightOn(int ln, int turnon, int newUIid = (-1));
};


//////////////////// highlight the Nth light
class CmdDisplayLightHL : public Command {
private:
  int n, hl;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayLightHL(int ln, int highlt, int newUIid = (-1));
};


//////////////////// rotate the position of the Nth light
class CmdDisplayLightRot : public Command {
private:
  float theta;
  char axis;
  int n;
protected:
  virtual int do_execute(void) ;
  virtual void create_text(void);
public:
  CmdDisplayLightRot(int ln, float th, char ax, int newUIid = (-1));
};


#endif

